use std::fmt::{Debug, Display};
use std::ops::Range;
use std::sync::Arc;

use logos::{Lexer, Logos};

use crate::args::{Arg, Args};
use crate::source::{Source, SourcePosition, SourceRange, WithSource};

#[derive(Logos, Copy, Clone, Debug, PartialEq, Eq)]
#[logos(subpattern ident = r"[a-zA-Z][a-zA-Z0-9]*")]
pub enum TokenKind {
    #[token("(")]
    ParenOpen,

    #[token(")")]
    ParenClose,

    #[regex(r"[a-zA-Z-][a-zA-Z0-9-]*")]
    Symbol,

    #[regex(r"-?[0-9]*")]
    Number,

    #[regex(":[a-zA-Z-][a-zA-Z0-9-]*")]
    Keyword,

    #[regex(r#""([^"\\]|\\t|\\u|\\n|\\")*""#)]
    String,

    #[regex(r";[^\n]*", logos::skip)]
    Comment,

    #[token("\n")]
    Newline,

    #[regex(r"[ \t\f\r]+", logos::skip)]
    Whitespace,

    #[logos(ignore)]
    Eof,

    #[error]
    Error,
}

#[derive(Debug, Clone)]
pub struct TokenInner {
    pub kind: TokenKind,
    pub content: String,
}

pub type Token = WithSource<TokenInner>;

#[derive(Debug)]
pub struct AstBuilder<'a> {
    lexer: Lexer<'a, TokenKind>,
    source: &'a Arc<Source>,
    current_row_idx: usize,
    current_row_start: usize,
    paren_level: usize,
}

impl<'a> AstBuilder<'a> {
    pub fn new(source: &'a Arc<Source>) -> Self {
        Self {
            lexer: TokenKind::lexer(&source.full),
            current_row_idx: 0,
            current_row_start: 0,
            paren_level: 0,
            source,
        }
    }

    pub fn idx_to_position(&self, idx: usize) -> SourcePosition {
        SourcePosition {
            row: self.current_row_idx,
            col: idx - self.current_row_start,
        }
    }

    pub fn span_to_range(&self, span: Range<usize>) -> SourceRange {
        Range {
            start: self.idx_to_position(span.start),
            end: self.idx_to_position(span.end),
        }
    }

    pub fn lexer_range(&self) -> SourceRange {
        self.span_to_range(self.lexer.span())
    }

    pub fn with_source<T>(&self, inner: T, range: SourceRange) -> WithSource<T> {
        WithSource {
            inner,
            source: self.source.clone(),
            range,
        }
    }

    pub fn token_to_string(&self, token: Token) -> ParseResult<WithSource<String>> {
        match token.kind {
            TokenKind::String => {
                let slice = self.lexer.slice();
                let end = slice.len() - 1;
                let inner = slice[1..end].to_string();
                Ok(WithSource {
                    inner,
                    source: token.source,
                    range: token.range,
                })
            }
            TokenKind::Symbol => Ok(WithSource {
                inner: token.content.clone(),
                source: token.source,
                range: token.range,
            }),
            TokenKind::Keyword => Ok(WithSource {
                inner: token.content[1..].to_string(),
                source: token.source,
                range: token.range,
            }),
            _ => Err(token.map(ParseErrorKind::UnexpectedToken(token.kind, None))),
        }
    }

    pub fn next(&mut self) -> ParseResult<Token> {
        loop {
            let token = self.lexer.next().unwrap_or(TokenKind::Eof);
            let span = self.lexer.span();
            if let TokenKind::Newline = token {
                self.current_row_idx += 1;
                self.current_row_start = span.end;
            } else if let TokenKind::Error = token {
                let range = self.span_to_range(span);
                break Err(self.with_source(ParseErrorKind::InvalidToken, range));
            } else {
                let range = self.span_to_range(span);

                if let TokenKind::ParenOpen = token {
                    self.paren_level += 1;
                }

                if let TokenKind::ParenClose = token {
                    if self.paren_level == 0 {
                        return Err(self.with_source(ParseErrorKind::UnmatchedParen, range));
                    }

                    self.paren_level -= 1;
                }

                if let TokenKind::Eof = token {
                    if self.paren_level > 0 {
                        return Err(self.with_source(ParseErrorKind::UnmatchedParen, range));
                    }
                }

                break Ok(Token {
                    inner: TokenInner {
                        kind: token,
                        content: self.lexer.slice().to_string(),
                    },
                    source: self.source.to_owned(),
                    range,
                });
            }
        }
    }

    pub fn expect_next(&mut self) -> ParseResult<Token> {
        let tok = self.next()?;
        if let TokenKind::Eof = tok.kind {
            Err(tok.map(ParseErrorKind::UnexpectedEof))
        } else {
            Ok(tok)
        }
    }

    pub fn parse_tag_body(&mut self) -> ParseResult<TagBody> {
        let mut args = Args::new();
        let children = loop {
            let token = self.next()?;
            match token.kind {
                TokenKind::Eof => break vec![],
                TokenKind::ParenClose => break vec![],
                TokenKind::Keyword => {
                    let name = self.token_to_string(token.clone())?;

                    if args.contains(&name) {
                        return Err(token.map(ParseErrorKind::DuplicateArgument(name.inner)));
                    }

                    let val = self.expect_value()?;
                    args.insert(Arg { name, val });
                }
                _ => break self.parse_list_from(token)?,
            }
        };

        Ok(TagBody { args, children })
    }

    pub fn parse_list_from(&mut self, mut token: Token) -> ParseResult<Vec<Value>> {
        let mut vals = Vec::new();

        loop {
            if let TokenKind::Symbol = token.kind {
                let start = token.range.start.clone();
                let id = self.token_to_string(token)?;
                let body = self.parse_tag_body()?;
                let end = self.idx_to_position(self.lexer.span().end);
                let val = Value {
                    inner: ValueKind::Tag(Tag { id, body }),
                    source: self.source.to_owned(),
                    range: start..end,
                };
                vals.push(val);
                break;
            } else if let Some(val) = self.parse_value_from(token)? {
                vals.push(val);
                token = self.next()?;
            } else {
                break;
            }
        }

        Ok(vals)
    }

    pub fn parse_value_from(&mut self, token: Token) -> ParseResult<Option<Value>> {
        match token.kind {
            TokenKind::Eof | TokenKind::ParenClose => Ok(None),
            TokenKind::String => Ok(Some(self.token_to_string(token)?.to_val_string())),
            TokenKind::Symbol => Ok(Some(self.token_to_string(token)?.to_val_symbol())),
            TokenKind::Number => {
                if let Ok(val) = token.content.parse() {
                    Ok(Some(token.map(ValueKind::Number(val))))
                } else {
                    Err(token.map(ParseErrorKind::BadNumber))
                }
            }
            TokenKind::ParenOpen => {
                let start = token.range.end;
                let next = self.expect_next()?;
                let vals = self.parse_list_from(next)?;
                let end = self.idx_to_position(self.lexer.span().end);
                Ok(Some(Value {
                    inner: ValueKind::List(vals),
                    source: self.source.to_owned(),
                    range: start..end,
                }))
            }
            _ => Err(token.map(ParseErrorKind::UnexpectedToken(token.kind, None))),
        }
    }

    pub fn expect_value(&mut self) -> ParseResult<Value> {
        let next = self.expect_next()?;
        self.parse_value_from(next)?
            .ok_or_else(|| self.with_source(ParseErrorKind::UnexpectedEof, self.lexer_range()))
    }
}

#[derive(Clone, Debug)]
pub enum ValueKind {
    String(String),
    Number(i32),
    Symbol(String),
    Tag(Tag),
    List(Vec<Value>),
}

pub type Value = WithSource<ValueKind>;

impl WithSource<String> {
    pub fn to_val_string(self) -> Value {
        Value {
            inner: ValueKind::String(self.inner),
            source: self.source,
            range: self.range,
        }
    }

    pub fn to_val_symbol(self) -> Value {
        Value {
            inner: ValueKind::Symbol(self.inner),
            source: self.source,
            range: self.range,
        }
    }
}

#[derive(Clone, Debug)]
pub struct Tag {
    pub id: WithSource<String>,
    pub body: TagBody,
}

#[derive(Clone, Debug, Default)]
pub struct TagBody {
    pub args: Args,
    pub children: Vec<Value>,
}

pub type ParseError = WithSource<ParseErrorKind>;
pub type ParseResult<T> = Result<T, ParseError>;

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ParseErrorKind {
    InvalidToken,
    DuplicateArgument(String),
    MissingArgument(String),
    UnmatchedParen,
    UnexpectedToken(TokenKind, Option<TokenKind>),
    UnexpectedEof,
    InvalidValue,
    BadNumber,
}

impl Display for ParseErrorKind {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "Parse error: ")?;

        use ParseErrorKind::*;
        match self {
            InvalidToken => write!(f, "invalid token"),
            DuplicateArgument(name) => write!(f, "duplicate {} argument", name),
            MissingArgument(name) => write!(f, "missing {} argument", name),
            UnmatchedParen => write!(f, "unmatched parentheses"),
            UnexpectedToken(kind, Some(expected)) => {
                write!(f, "unexpected {:?} token (expected {:?})", kind, expected)
            }
            UnexpectedToken(kind, None) => write!(f, "unexpected token {:?}", kind),
            UnexpectedEof => write!(f, "unexpected end-of-file"),
            InvalidValue => write!(f, "invalid value"),
            BadNumber => write!(f, "badly-formed number"),
        }
    }
}
