use strum::EnumString;

use crate::args::{Args, ParseArgs, ParseScoped};
use crate::ast::ParseResult;
use crate::border::BorderKind;
use crate::layout::{AxisSize, Dimensions, Margin, Rect, Axis};
use crate::text::{Color, TextLayout};

#[derive(Copy, Clone, Debug, Hash, PartialEq, Eq, EnumString)]
#[strum(serialize_all = "kebab-case", ascii_case_insensitive)]
pub enum BlockTagKind {
    Section,
    P,
    Figlet,
    Code,
    Ul,
    Ol,
    Li,
    Box,
}

impl BlockTagKind {
    pub fn children(&self) -> AllowedChildren {
        use BlockTagKind::*;
        match self {
            Section => AllowedChildren::Block,
            P => AllowedChildren::TaggedText,
            Figlet | Code => AllowedChildren::PlainText,
            Ul | Ol => AllowedChildren::Specific(&[Li, Ul, Ol]),
            Li => AllowedChildren::TaggedText,
            Box => AllowedChildren::Block,
        }
    }

    pub fn make_text_layout(&self, text: String) -> TextLayout {
        use BlockTagKind::*;
        match self {
            Figlet => {
                let font = figlet_rs::FIGfont::standard().unwrap();
                let figure = font.convert(&text).expect("Failed to FIG-ify text");
                TextLayout::from_plain(&figure.to_string())
            }
            Code => TextLayout::from_plain(&text),
            _ => unimplemented!("Text layout for {:?} tag", self),
        }
    }
}

#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub enum AllowedChildren {
    Block,
    PlainText,
    TaggedText,
    Specific(&'static [BlockTagKind]),
}

impl AllowedChildren {
    pub fn allows_styled_text(&self) -> bool {
        use AllowedChildren::*;
        match self {
            TaggedText => true,
            _ => false,
        }
    }

    pub fn allows_block(&self, kind: &BlockTagKind) -> bool {
        use AllowedChildren::*;
        match self {
            Block => true,
            Specific(allowed) => allowed.contains(kind),
            _ => false,
        }
    }
}

#[derive(Clone, Debug)]
pub struct Block {
    pub size: Dimensions<AxisSize>,
    pub margin: Rect<Margin>,
    pub padding: Rect<usize>,
    pub border: Rect<Option<BorderKind>>,
    pub border_color: Rect<Color>,
}

impl ParseArgs for Block {
    fn parse_args(args: &mut Args) -> ParseResult<Self> {
        let size = Dimensions::parse(args, 1)?;
        let margin = Rect::parse_scoped(args, "margin", Margin::Fixed(1))?;
        let padding = Rect::parse_scoped(args, "padding", 0)?;
        let border = Rect::parse_scoped(args, "border", None)?;
        let border_color = Rect::parse_scoped(args, "border-color", Color::White)?;

        Ok(Self {
            size,
            margin,
            padding,
            border,
            border_color,
        })
    }
}

impl Block {
    pub fn required_axis_size(&self, axis: Axis) -> AxisSize {
        let mut size = self.size.axis(axis).to_owned();

        let (ps, pe) = self.padding.axis_sides(axis);
        size.add_size(*ps);
        size.add_size(*pe);

        let (ms, me) = self.margin.axis_sides(axis);
        size.add_margin(ms);
        size.add_margin(me);

        size
    }
}
