use std::ops::{Deref, DerefMut};
use std::str::FromStr;

use crate::args::{Args, ParseArgs, ParseScoped};
use crate::ast::{self, ParseErrorKind, ParseResult, TagBody};
use crate::block::{AllowedChildren, Block, BlockTagKind};
use crate::layout::{Axis, BoxLayout, Dimensions, RenderedBlock};
use crate::node::{NodeId, NodeStore};
use crate::source::WithSource;
use crate::text::{TaggedText, TextBlock, TextLayout, TextParseErrorKind, TextStyle};

#[derive(Clone, Debug)]
pub struct Dom {
    pub root: NodeId,
    pub nodes: NodeStore,
}

impl Dom {
    pub fn parse(body: &TagBody) -> DomResult<Self> {
        let mut builder = DomBuilder::default();
        let root = builder.new_node();
        let mut block = BoxBuilder::root(root);
        let mut text_style = TextStyle::default();

        let mut state = DomWalkState {
            builder: &mut builder,
            box_builder: &mut block,
            text_style: &mut text_style,
        };

        state.add_children(&body.children)?;

        block.build(&mut builder);

        Ok(Self {
            root,
            nodes: builder.node_store,
        })
    }

    pub fn print(&self) {
        self.print_from(self.root, 0);
    }

    pub fn print_from(&self, node: NodeId, indent: usize) {
        let margin: String = std::iter::repeat(" ").take(indent * 4).collect();
        if let Some(NodeChildren(children)) = self.nodes.get(node) {
            println!("{}{}: {:?}", margin, node, children);

            for child in children.iter() {
                self.print_from(*child, indent + 1);
            }
        } else {
            let text = self.nodes.get::<TextBlock>(node);
            println!("{}{}: {:?}", margin, node, text);
        }
    }

    pub fn render(&mut self, size: Dimensions<usize>) -> &RenderedBlock {
        self.render_from(self.root, size);
        self.nodes.get(self.root).unwrap()
    }

    pub fn render_from(&mut self, node: NodeId, size: Dimensions<usize>) {
        let rendered = if let Some(NodeChildren(children)) = self.nodes.get(node) {
            let children = children.to_owned();
            let axis: Axis = self.nodes.get::<Axis>(node).unwrap().to_owned();

            let layout = BoxLayout {
                axis,
                children: self.get_many::<Block>(&children),
            };

            let sizes = layout.calc_sizes(size).to_owned();
            drop(layout);

            for (size, child) in sizes.into_iter().zip(children) {
                self.render_from(child, size);
            }

            let children = self.get_children(node);
            let rendered = self.get_many::<RenderedBlock>(&children);

            let layout = BoxLayout {
                axis,
                children: self.get_many::<Block>(&children),
            };

            layout.render(&rendered)
        } else {
            let style = TextStyle::default();

            let layout: TextLayout = if let Some(layout) = self.nodes.get::<TextLayout>(node) {
                layout.to_owned()
            } else if let Some(block) = self.nodes.get::<TextBlock>(node) {
                block.layout(&style, size.width)
            } else {
                panic!("Couldn't find text info on node {}", node);
            };

            let rendered = layout.render(&style);
            let block: &Block = self.nodes.get(node).unwrap();
            rendered.add_padding(&block.padding)
        };

        self.nodes.insert(node, rendered);
    }

    pub fn get_children(&self, node: NodeId) -> Vec<NodeId> {
        self.nodes.get::<NodeChildren>(node).unwrap().0.to_owned()
    }

    pub fn get_many<T: 'static>(&self, children: &[NodeId]) -> Vec<&T> {
        children
            .iter()
            .map(|child| self.nodes.get(*child).unwrap())
            .collect()
    }
}

#[derive(Clone, Debug, Default)]
pub struct NodeChildren(Vec<NodeId>);

#[derive(Copy, Clone, Debug, Default)]
pub struct NodeParent(NodeId);

#[derive(Debug, Default)]
pub struct DomBuilder {
    pub node_store: NodeStore,
    pub next_node: NodeId,
}

impl Deref for DomBuilder {
    type Target = NodeStore;
    fn deref(&self) -> &NodeStore {
        &self.node_store
    }
}

impl DerefMut for DomBuilder {
    fn deref_mut(&mut self) -> &mut NodeStore {
        &mut self.node_store
    }
}

impl DomBuilder {
    pub fn new_node(&mut self) -> NodeId {
        let node = self.next_node;
        self.next_node += 1;
        node
    }
}

#[derive(Debug)]
pub struct BoxBuilder {
    pub id: NodeId,
    pub axis: Axis,
    pub children: Vec<NodeId>,
}

impl BoxBuilder {
    pub fn new(id: NodeId, args: &mut Args) -> ParseResult<Self> {
        Ok(Self {
            id,
            axis: Axis::parse_scoped(args, "axis", Axis::Vertical)?,
            children: Vec::new(),
        })
    }

    pub fn root(id: NodeId) -> Self {
        Self {
            id,
            axis: Axis::Vertical,
            children: Vec::new(),
        }
    }

    pub fn build(self, dom: &mut DomBuilder) -> NodeId {
        for child in self.children.iter() {
            dom.insert(*child, NodeParent(self.id));
        }

        let children = NodeChildren(self.children);
        dom.insert(self.id, children);
        dom.insert(self.id, self.axis);

        self.id
    }
}

#[derive(Debug)]
pub struct DomWalkState<'a> {
    pub builder: &'a mut DomBuilder,
    pub box_builder: &'a mut BoxBuilder,
    pub text_style: &'a mut TextStyle,
}

impl<'a> DomWalkState<'a> {
    pub fn fork(&mut self) -> DomWalkState {
        DomWalkState {
            builder: &mut *self.builder,
            box_builder: &mut *self.box_builder,
            text_style: &mut *self.text_style,
        }
    }

    pub fn add_text_block(&mut self, kind: BlockTagKind, text: TaggedText) -> NodeId {
        let node = self.builder.new_node();
        let block = text.make_text_block(kind);
        self.builder.insert(node, block);
        node
    }

    pub fn add_block_tag(&mut self, kind: BlockTagKind, body: &ast::TagBody) -> DomResult<NodeId> {
        let node = self.builder.new_node();

        let mut args = body.args.clone();
        let block = Block::parse_args(&mut args).map_err(|e| e.map(e.inner.clone().into()))?;
        println!("{:#?}", block);
        self.builder.insert(node, block);

        if let BlockTagKind::Box = kind {
            self.box_builder.children.push(node);
            let mut box_builder =
                BoxBuilder::new(node, &mut args).map_err(|e| e.map(e.inner.clone().into()))?;
            let mut state = DomWalkState {
                box_builder: &mut box_builder,
                ..self.fork()
            };

            state.add_children(&body.children)?;
            box_builder.build(&mut self.builder);
            return Ok(node);
        }

        match kind.children() {
            AllowedChildren::TaggedText => {
                let text = TaggedText::parse(&self.text_style, &body.children)
                    .map_err(|e| e.map(e.inner.clone().into()))?;
                let block = text.make_text_block(kind);
                self.builder.insert(node, block);
                self.box_builder.children.push(node);
            }
            AllowedChildren::PlainText => {
                let text = crate::text::parse_plain(&body.children)
                    .map_err(|e| e.map(e.inner.clone().into()))?;
                let layout = kind.make_text_layout(text);
                self.builder.insert(node, layout);
                self.box_builder.children.push(node);
            }
            _ => {
                self.add_children(&body.children)?;
            }
        }

        Ok(node)
    }

    pub fn add_tag(mut self, tag: &ast::Tag) -> DomResult<()> {
        if let Ok(kind) = BlockTagKind::from_str(&tag.id) {
            self.add_block_tag(kind, &tag.body)?;
            Ok(())
        } else {
            Err(tag.id.map(DomErrorKind::InvalidTag))
        }
    }

    pub fn add_children(&mut self, vals: &[ast::Value]) -> DomResult<()> {
        for child in vals.iter() {
            self.add_child(child)?;
        }

        Ok(())
    }

    pub fn add_child(&mut self, val: &ast::Value) -> DomResult<()> {
        use ast::ValueKind::*;
        match &val.inner {
            Tag(tag) => {
                let state = self.fork();
                state.add_tag(tag)
            }
            List(list) => self.add_children(list),
            String(_string) => {
                println!("{}", val.map("Insert string code here:"));
                Ok(())
            }
            _ => Err(val.map(DomErrorKind::InvalidChild)),
        }
    }
}

pub type DomResult<T> = Result<T, WithSource<DomErrorKind>>;

#[derive(Clone, Debug)]
pub enum DomErrorKind {
    InvalidTag,
    InvalidChild,
    ParseError(ParseErrorKind),
    TextError(TextParseErrorKind),
}

impl From<ParseErrorKind> for DomErrorKind {
    fn from(e: ParseErrorKind) -> Self {
        DomErrorKind::ParseError(e)
    }
}

impl From<TextParseErrorKind> for DomErrorKind {
    fn from(e: TextParseErrorKind) -> Self {
        DomErrorKind::TextError(e)
    }
}

impl std::fmt::Display for DomErrorKind {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        use DomErrorKind::*;
        if let TextError(e) = &self {
            return e.fmt(f);
        }

        if let ParseError(e) = &self {
            return e.fmt(f);
        }

        write!(f, "DOM error: ")?;

        match self {
            InvalidTag => write!(f, "invalid tag"),
            InvalidChild => write!(f, "invalid child"),
            _ => unimplemented!(),
        }
    }
}
