use itertools::Itertools;
use strum::EnumString;

use crate::args::*;
use crate::ast::{ParseErrorKind, ParseResult, Value, ValueKind};
use crate::block::Block;
use crate::source::WithSource;

#[derive(Clone, Debug)]
pub struct RenderedBlock {
    pub lines: Vec<String>,
    pub size: Dimensions<usize>,
}

impl RenderedBlock {
    pub fn add_padding(&self, padding: &Rect<usize>) -> Self {
        let make_spaces = |num: usize| -> String { std::iter::repeat(" ").take(num).collect() };
        let new_width = self.size.width + padding.l + padding.r;
        let left_pad = make_spaces(padding.l);
        let right_pad = make_spaces(padding.r);
        let vertical_pad = make_spaces(new_width);

        let mut lines = Vec::new();

        lines.extend(std::iter::repeat(vertical_pad.clone()).take(padding.t));

        lines.extend(
            self.lines
                .iter()
                .map(|line| format!("{}{}{}", left_pad, line, right_pad)),
        );

        lines.extend(std::iter::repeat(vertical_pad.clone()).take(padding.b));

        Self {
            lines,
            size: Dimensions {
                width: new_width,
                height: self.size.height + padding.t + padding.b,
            },
        }
    }

    pub fn make_blank(size: Dimensions<usize>) -> Self {
        let line: String = std::iter::repeat(" ").take(size.width).collect();
        Self {
            lines: std::iter::repeat(line).take(size.height).collect(),
            size,
        }
    }

    pub fn concat(mut self, axis: Axis, mut other: Self) -> Self {
        match axis {
            Axis::Vertical => Self {
                size: Dimensions {
                    width: self.size.width.max(other.size.width),
                    height: self.size.height + other.size.height,
                },
                lines: self.lines.into_iter().chain(other.lines).collect(),
            },
            Axis::Horizontal => {
                if self.size.height < other.size.height {
                    let padding = Dimensions {
                        width: self.size.width,
                        height: other.size.height - self.size.height,
                    };

                    let padding = Self::make_blank(padding);
                    self = self.concat(Axis::Vertical, padding);
                }

                if other.size.height < self.size.height {
                    let padding = Dimensions {
                        width: other.size.width,
                        height: self.size.height - other.size.height,
                    };

                    let padding = Self::make_blank(padding);
                    other = other.concat(Axis::Vertical, padding);
                }

                Self {
                    size: Dimensions {
                        width: self.size.width + other.size.width,
                        height: self.size.height.max(other.size.height),
                    },
                    lines: self
                        .lines
                        .into_iter()
                        .zip(other.lines)
                        .map(|(first, other)| format!("{}{}", first, other))
                        .collect(),
                }
            }
        }
    }
}

#[derive(Copy, Clone, Debug, EnumString)]
#[strum(serialize_all = "kebab-case", ascii_case_insensitive)]
pub enum Axis {
    Horizontal,
    Vertical,
}

impl ParseSymbol for Axis {}

impl Axis {
    pub fn cross(&self) -> Self {
        match self {
            Axis::Horizontal => Axis::Vertical,
            Axis::Vertical => Axis::Horizontal,
        }
    }
}

#[derive(Clone, Debug)]
pub struct BoxLayout<'a> {
    pub axis: Axis,
    pub children: Vec<&'a Block>,
}

impl<'a> BoxLayout<'a> {
    pub fn calc_sizes(&self, available: Dimensions<usize>) -> Vec<Dimensions<usize>> {
        let cross = self.axis.cross();

        let sizes = self
            .children
            .iter()
            .map(|block| {
                let available = *available.axis(cross);
                let required = block.required_axis_size(cross);
                let min_size = required.min_size();
                let remaining = available - min_size.min(available);
                let cross_size = block.size.axis(cross).clamp(remaining);
                let axis_size = block.size.axis(self.axis).min_size();

                match self.axis {
                    Axis::Horizontal => Dimensions {
                        width: axis_size,
                        height: cross_size,
                    },
                    Axis::Vertical => Dimensions {
                        width: cross_size,
                        height: axis_size,
                    },
                }
            })
            .collect();

        sizes
    }

    pub fn render(&self, blocks: &[&RenderedBlock]) -> RenderedBlock {
        let axis = self.axis;
        let cross = axis.cross();

        let cross_size = blocks
            .iter()
            .zip(&self.children)
            .map(|(rendered, block)| {
                let (ms, me) = block.margin.axis_sides(cross);
                ms.required_size() + rendered.size.axis(cross) + me.required_size()
            })
            .max()
            .unwrap_or(0);

        let make_spacer = |size: usize| -> RenderedBlock {
            let (width, height) = match axis {
                Axis::Horizontal => (size, cross_size),
                Axis::Vertical => (cross_size, size),
            };

            RenderedBlock::make_blank(Dimensions { width, height })
        };

        let mut spacers = Vec::new();
        let mut last_margin = 0;

        for block in self.children.iter() {
            let (ms, me) = block.margin.axis_sides(axis);
            let size = ms.required_size().max(last_margin);
            spacers.push(make_spacer(size));
            last_margin = me.required_size();
        }

        spacers.push(make_spacer(last_margin));

        let mut padded = Vec::new();

        for (block, rendered) in self.children.iter().zip(blocks) {
            let (ms, me) = block.margin.axis_sides(cross);
            let available = cross_size - rendered.size.axis(cross).min(&cross_size);
            let (ms, me) = Self::calc_margins(ms, me, available);

            let padding = match axis {
                Axis::Horizontal => Rect {
                    l: 0,
                    r: 0,
                    t: ms,
                    b: me,
                },
                Axis::Vertical => Rect {
                    t: 0,
                    b: 0,
                    l: ms,
                    r: me,
                },
            };

            padded.push(rendered.add_padding(&padding));
        }

        spacers
            .into_iter()
            .interleave(padded)
            .reduce(|r1, r2| r1.concat(axis, r2))
            .unwrap()
    }

    pub fn calc_margins(start: &Margin, end: &Margin, available: usize) -> (usize, usize) {
        let split_available = |available: usize| -> (usize, usize) {
            if available & 2 == 0 {
                (available >> 1, available >> 1)
            } else {
                (available >> 1, (available >> 1) | 1)
            }
        };

        use Margin::*;
        match (start, end) {
            (Auto, Auto) => split_available(available),
            (Fixed(start), Auto) => (*start, available - start),
            (Auto, Fixed(end)) => (available - end, *end),
            (Fixed(fixed_start), Fixed(fixed_end)) => {
                let taken = fixed_start + fixed_end;
                let available = available - taken.min(available);
                let (start, end) = split_available(available);
                (fixed_start + start, fixed_end + end)
            }
        }
    }
}

#[derive(Clone, Debug)]
pub enum AxisSize {
    Fixed(usize),
    Bounded(usize, Option<usize>),
}

impl ParseScoped for AxisSize {
    type Default = usize;

    fn parse_scoped(args: &mut Args, name: &str, mut min: usize) -> ParseResult<Self> {
        if let Some(arg) = args.get(name) {
            Ok(AxisSize::Fixed(ParseValue::parse_value(&arg.val)?))
        } else {
            let min_name = format!("min-{}", name);
            if let Some(arg) = args.get(&min_name) {
                min = ParseValue::parse_value(&arg.val)?;
            }

            let max_name = format!("max-{}", name);
            let max = if let Some(arg) = args.get(&max_name) {
                Some(ParseValue::parse_value(&arg.val)?)
            } else {
                None
            };

            Ok(AxisSize::Bounded(min, max))
        }
    }
}

impl AxisSize {
    pub fn add_size(&mut self, size: usize) {
        match self {
            AxisSize::Fixed(fixed) => *fixed += size,
            AxisSize::Bounded(min, None) => *min += size,
            AxisSize::Bounded(min, Some(max)) => {
                *min += size;
                *max += size;
            }
        }
    }

    pub fn add_margin(&mut self, margin: &Margin) {
        match margin {
            Margin::Fixed(margin) => self.add_size(*margin),
            Margin::Auto => {}
        }
    }

    pub fn clamp(&self, size: usize) -> usize {
        match *self {
            AxisSize::Fixed(fixed) => fixed,
            AxisSize::Bounded(min, None) => size.max(min),
            AxisSize::Bounded(min, Some(max)) => size.max(min).min(max),
        }
    }

    pub fn min_size(&self) -> usize {
        match self {
            AxisSize::Fixed(fixed) => *fixed,
            AxisSize::Bounded(min, _) => *min,
        }
    }
}

#[derive(Clone, Debug)]
pub enum Margin {
    Auto,
    Fixed(usize),
}

impl ParseValue for Margin {
    fn parse_value(val: &Value) -> ParseResult<Self> {
        match &val.inner {
            ValueKind::Symbol(symbol) if symbol.to_lowercase() == "auto" => Ok(Margin::Auto),
            _ => Ok(Margin::Fixed(usize::parse_value(val)?)),
        }
    }
}

impl Margin {
    pub fn required_size(&self) -> usize {
        match self {
            Margin::Auto => 0,
            Margin::Fixed(fixed) => *fixed,
        }
    }
}

#[derive(Copy, Clone, Debug)]
pub struct Dimensions<T> {
    pub width: T,
    pub height: T,
}

impl<T: ParseScoped> Dimensions<T>
where
    T::Default: Clone,
{
    pub fn parse(args: &mut Args, default: T::Default) -> ParseResult<Self> {
        let width = T::parse_scoped(args, "width", default.clone())?;
        let height = T::parse_scoped(args, "height", default)?;
        Ok(Self { width, height })
    }
}

impl<T> Dimensions<T> {
    pub fn axis(&self, axis: Axis) -> &T {
        match axis {
            Axis::Horizontal => &self.width,
            Axis::Vertical => &self.height,
        }
    }
}

impl Dimensions<AxisSize> {
    pub fn min_size(&self) -> Dimensions<usize> {
        Dimensions {
            width: self.width.min_size(),
            height: self.height.min_size(),
        }
    }
}

#[derive(Copy, Clone, Debug)]
pub struct Rect<T> {
    pub l: T,
    pub t: T,
    pub r: T,
    pub b: T,
}

impl<T> Rect<T> {
    pub fn map<O>(&self, f: impl Fn(&T) -> O) -> Rect<O> {
        Rect::<O> {
            l: f(&self.l),
            t: f(&self.t),
            r: f(&self.r),
            b: f(&self.b),
        }
    }

    pub fn axis_sides(&self, axis: Axis) -> (&T, &T) {
        match axis {
            Axis::Horizontal => (&self.l, &self.r),
            Axis::Vertical => (&self.t, &self.b),
        }
    }
}

impl<T: Clone + ParseValue> Rect<T> {
    pub fn parse_shorthand(parsed: WithSource<&[T]>) -> ParseResult<Self> {
        match parsed.len() {
            1 => Ok(Self {
                l: parsed[0].clone(),
                t: parsed[0].clone(),
                r: parsed[0].clone(),
                b: parsed[0].clone(),
            }),
            2 => Ok(Self {
                t: parsed[0].clone(),
                b: parsed[0].clone(),
                l: parsed[1].clone(),
                r: parsed[1].clone(),
            }),
            3 => Ok(Self {
                t: parsed[0].clone(),
                l: parsed[1].clone(),
                r: parsed[1].clone(),
                b: parsed[2].clone(),
            }),
            4 => Ok(Self {
                t: parsed[0].clone(),
                r: parsed[1].clone(),
                b: parsed[2].clone(),
                l: parsed[3].clone(),
            }),
            _ => Err(parsed.map(ParseErrorKind::InvalidValue)),
        }
    }
}

impl<T: Clone + ParseValue> ParseScoped for Rect<T> {
    type Default = T;

    fn parse_scoped(args: &mut Args, name: &str, default: T) -> ParseResult<Self> {
        let mut get_side = |side: &str| -> Option<Value> {
            let name = format!("{}-{}", name, side);
            args.get(&name).map(|arg| arg.val)
        };

        let l = get_side("left");
        let t = get_side("top");
        let r = get_side("right");
        let b = get_side("bottom");

        if let Some(short) = args.get(name) {
            if l.is_some() || t.is_some() || r.is_some() || b.is_some() {
                let e = ParseErrorKind::DuplicateArgument(name.to_string());
                return Err(short.name.map(e));
            }

            match &short.val.inner {
                ValueKind::List(vals) => {
                    let mut parsed = Vec::new();

                    for val in vals.iter() {
                        parsed.push(T::parse_value(val)?);
                    }

                    Self::parse_shorthand(short.val.map(&parsed))
                }
                _ => {
                    let val = T::parse_value(&short.val)?;
                    Self::parse_shorthand(short.val.map(&[val]))
                }
            }
        } else {
            let parse_value = |val: Option<Value>| -> ParseResult<T> {
                if let Some(val) = val {
                    T::parse_value(&val)
                } else {
                    Ok(default.clone())
                }
            };

            Ok(Self {
                l: parse_value(l)?,
                t: parse_value(t)?,
                r: parse_value(r)?,
                b: parse_value(b)?,
            })
        }
    }
}
