use std::any::{Any, TypeId};
use std::fmt::Debug;

use fnv::FnvHashMap as HashMap;

pub type NodeId = u32;
pub type ComponentStorage<T> = HashMap<NodeId, T>;

#[derive(Debug, Default)]
pub struct NodeStore {
    storage: HashMap<TypeId, Box<dyn ComponentStorageTrait>>,
}

impl NodeStore {
    pub fn get_storage<T: 'static>(&self) -> Option<&ComponentStorage<T>> {
        let id = TypeId::of::<T>();
        let boxed = self.storage.get(&id)?;
        boxed.as_any().downcast_ref()
    }

    pub fn get_storage_mut<T: 'static>(&mut self) -> Option<&mut ComponentStorage<T>> {
        let id = TypeId::of::<T>();
        let boxed = self.storage.get_mut(&id)?;
        boxed.as_any_mut().downcast_mut()
    }

    pub fn get<T: 'static>(&self, id: NodeId) -> Option<&T> {
        let storage = self.get_storage::<T>()?;
        storage.get(&id)
    }

    pub fn get_mut<T: 'static>(&mut self, id: NodeId) -> Option<&mut T> {
        let storage = self.get_storage_mut::<T>()?;
        storage.get_mut(&id)
    }

    pub fn insert<T: 'static + Debug + Clone>(&mut self, id: NodeId, component: T) -> Option<T> {
        if let Some(storage) = self.get_storage_mut::<T>() {
            storage.insert(id, component)
        } else {
            let mut storage = ComponentStorage::<T>::default();
            let old = storage.insert(id, component);
            let id = TypeId::of::<T>();
            self.storage.insert(id, Box::new(storage));
            old
        }
    }
}

impl Clone for NodeStore {
    fn clone(&self) -> Self {
        let mut storage = HashMap::default();

        for (id, src) in self.storage.iter() {
            storage.insert(*id, src.as_ref().clone());
        }

        Self { storage }
    }
}

pub trait ComponentStorageTrait: Any + Debug {
    fn clone(&self) -> Box<dyn ComponentStorageTrait + 'static>;
    fn as_any(&self) -> &dyn Any;
    fn as_any_mut(&mut self) -> &mut dyn Any;
}

impl<T: Clone + Debug + 'static> ComponentStorageTrait for ComponentStorage<T> {
    fn clone(&self) -> Box<dyn ComponentStorageTrait + 'static> {
        Box::new(Clone::clone(self))
    }

    fn as_any(&self) -> &dyn Any {
        self as _
    }

    fn as_any_mut(&mut self) -> &mut dyn Any {
        self as _
    }
}
