use std::fmt::{Debug, Display};
use std::ops::{Deref, Range};
use std::sync::Arc;

#[derive(Clone)]
pub struct WithSource<T> {
    pub inner: T,
    pub source: Arc<Source>,
    pub range: SourceRange,
}

impl<T: Debug> Debug for WithSource<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        self.inner.fmt(f)
    }
}

impl<T: Debug + Display> Display for WithSource<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        writeln!(f, "  {}", self.inner)?;

        for line_idx in self.range.start.row..=self.range.end.row {
            let line = &self.source.lines[line_idx];
            writeln!(f, "  {:<4}| {}", line_idx + 1, line)?;

            let start = if line_idx == self.range.start.row {
                self.range.start.col
            } else {
                0
            };

            let end = if line_idx == self.range.end.row {
                self.range.end.col
            } else {
                line.len()
            };

            let spaces: String = std::iter::repeat(" ").take(start).collect();
            let carats: String = std::iter::repeat("^").take(end - start).collect();
            writeln!(f, "      | {}{}", spaces, carats)?;
        }

        Ok(())
    }
}

impl<T> AsRef<T> for WithSource<T> {
    fn as_ref(&self) -> &T {
        &self.inner
    }
}

impl<T> Deref for WithSource<T> {
    type Target = T;

    fn deref(&self) -> &T {
        &self.inner
    }
}

impl<T> WithSource<T> {
    pub fn map<O>(&self, other: O) -> WithSource<O> {
        WithSource {
            inner: other,
            source: self.source.to_owned(),
            range: self.range.clone(),
        }
    }
}

#[derive(Clone, Debug)]
pub struct Source {
    pub full: String,
    pub lines: Vec<String>,
}

impl Source {
    pub fn from_string(full: String) -> Self {
        Self {
            lines: full.lines().map(ToString::to_string).collect(),
            full,
        }
    }
}

#[derive(Clone, Debug, Hash, PartialEq, Eq)]
pub struct SourcePosition {
    pub row: usize,
    pub col: usize,
}

pub type SourceRange = Range<SourcePosition>;
