use std::str::FromStr;

use strum::EnumString;

use crate::ast;
use crate::block::BlockTagKind;
use crate::layout::{RenderedBlock, Dimensions};
use crate::source::WithSource;

#[derive(Clone, Debug, Hash, PartialEq, Eq, EnumString)]
#[strum(serialize_all = "kebab-case", ascii_case_insensitive)]
pub enum TextTagKind {
    Black,
    Red,
    Green,
    Yellow,
    Blue,
    Magenta,
    Cyan,
    White,
    Italic,
    Bold,
    Underline,
    Strike,
    Style,
}

impl TextTagKind {
    pub fn to_color(&self) -> Option<Color> {
        use Color as C;
        use TextTagKind::*;
        match self {
            Black => Some(C::Black),
            Red => Some(C::Red),
            Green => Some(C::Green),
            Yellow => Some(C::Yellow),
            Blue => Some(C::Blue),
            Magenta => Some(C::Magenta),
            Cyan => Some(C::Cyan),
            White => Some(C::White),
            _ => None,
        }
    }
}

#[derive(Copy, Clone, Debug, Hash, PartialEq, Eq, EnumString)]
#[strum(serialize_all = "kebab-case", ascii_case_insensitive)]
pub enum Color {
    Black,
    Red,
    Green,
    Yellow,
    Blue,
    Magenta,
    Cyan,
    White,
}

impl crate::args::ParseSymbol for Color {}

impl From<Color> for ansi_term::Color {
    fn from(c: Color) -> Self {
        use ansi_term::Color as C;
        use Color::*;
        match c {
            Black => C::Black,
            Red => C::Red,
            Green => C::Green,
            Yellow => C::Yellow,
            Blue => C::Blue,
            Magenta => C::Purple,
            Cyan => C::Cyan,
            White => C::White,
        }
    }
}

bitflags::bitflags! {
    pub struct TextStyleFlags: u32 {
        const ITALIC = 1 << 0;
        const BOLD = 1 << 1;
        const UNDERLINE = 1 << 2;
        const STRIKETHROUGH = 1 << 3;
    }
}

#[derive(Clone, Debug, Hash, PartialEq, Eq)]
pub struct TextStyle {
    pub fg: Option<Color>,
    pub bg: Option<Color>,
    pub flags: TextStyleFlags,
}

impl Default for TextStyle {
    fn default() -> Self {
        Self {
            fg: None,
            bg: None,
            flags: TextStyleFlags::empty(),
        }
    }
}

impl TextStyle {
    pub fn apply_tag(&mut self, kind: &TextTagKind) {
        if let Some(color) = kind.to_color() {
            self.fg = Some(color);
            return;
        }

        use TextStyleFlags as F;
        use TextTagKind::*;
        let flags = match kind {
            Italic => F::ITALIC,
            Bold => F::BOLD,
            Strike => F::STRIKETHROUGH,
            Underline => F::UNDERLINE,
            _ => unreachable!(),
        };

        self.flags |= flags;
    }

    pub fn inherit(&self, base: &TextStyle) -> Self {
        Self {
            fg: self.fg.or(base.fg),
            bg: self.bg.or(base.bg),
            flags: self.flags | base.flags,
        }
    }

    pub fn make_painter(&self) -> ansi_term::Style {
        ansi_term::Style {
            foreground: self.fg.map(Into::into),
            background: self.bg.map(Into::into),
            is_bold: self.flags.contains(TextStyleFlags::BOLD),
            is_italic: self.flags.contains(TextStyleFlags::ITALIC),
            is_underline: self.flags.contains(TextStyleFlags::UNDERLINE),
            is_strikethrough: self.flags.contains(TextStyleFlags::STRIKETHROUGH),
            ..Default::default()
        }
    }
}

#[derive(Clone, Debug, Hash, PartialEq, Eq)]
pub struct StyledString {
    pub string: String,
    pub style: TextStyle,
}

impl StyledString {
    pub fn from_plain(plain: &str) -> Self {
        Self {
            string: plain.to_string(),
            style: TextStyle::default(),
        }
    }

    pub fn render(&self, base: &TextStyle) -> String {
        let painter = self.style.inherit(base).make_painter();
        painter.paint(&self.string).to_string()
    }
}

#[derive(Clone, Debug, Default, Hash, PartialEq, Eq)]
pub struct TaggedText {
    pub strings: Vec<StyledString>,
}

impl TaggedText {
    pub fn from_plain(plain: &str) -> Self {
        Self {
            strings: vec![StyledString::from_plain(plain)],
        }
    }

    pub fn concat(&mut self, mut other: Self) {
        self.strings.append(&mut other.strings);
    }

    pub fn parse(style: &TextStyle, vals: &[ast::Value]) -> Result<Self, TextParseError> {
        let mut text = Self::default();
        for val in vals.iter() {
            match &val.inner {
                ast::ValueKind::String(string) => text.strings.push(StyledString {
                    string: string.clone(),
                    style: style.clone(),
                }),
                ast::ValueKind::List(items) => text.concat(Self::parse(style, items)?),
                ast::ValueKind::Tag(tag) => text.concat(Self::parse_tag(style, tag)?),
                _ => return Err(val.map(TextParseErrorKind::UnexpectedValue)),
            }
        }

        Ok(text)
    }

    pub fn parse_tag(style: &TextStyle, tag: &ast::Tag) -> Result<Self, TextParseError> {
        let kind = TextTagKind::from_str(&tag.id)
            .map_err(|_| tag.id.map(TextParseErrorKind::InvalidTag))?;
        Self::parse_tag_body(style, &kind, &tag.body)
    }

    pub fn parse_tag_body(
        style: &TextStyle,
        kind: &TextTagKind,
        body: &ast::TagBody,
    ) -> Result<Self, TextParseError> {
        let mut style = style.clone();
        style.apply_tag(kind);
        Self::parse(&style, &body.children)
    }

    pub fn render(&self, base: &TextStyle) -> String {
        let mut rendered = String::new();

        for string in self.strings.iter() {
            let string = string.render(base);
            rendered.push_str(&string);
        }

        rendered
    }

    pub fn make_text_block(&self, kind: BlockTagKind) -> TextBlock {
        use BlockTagKind::*;
        match kind {
            P => TextBlock {
                text: self.to_owned(),
                initial_indent: None,
                subsequent_indent: None,
            },
            _ => unimplemented!("Text block for {:?} tag", kind),
        }
    }
}

pub fn parse_plain(vals: &[ast::Value]) -> Result<String, TextParseError> {
    let mut text = String::new();
    for val in vals.iter() {
        match &val.inner {
            ast::ValueKind::List(vals) => text.push_str(&parse_plain(vals)?),
            ast::ValueKind::String(string) => text.push_str(&string),
            ast::ValueKind::Tag(_) => return Err(val.map(TextParseErrorKind::UnexpectedTag)),
            _ => return Err(val.map(TextParseErrorKind::UnexpectedValue)),
        }
    }

    Ok(text)
}

pub type TextParseError = WithSource<TextParseErrorKind>;

#[derive(Clone, Debug)]
pub enum TextParseErrorKind {
    InvalidTag,
    UnexpectedValue,
    UnexpectedTag,
}

impl std::fmt::Display for TextParseErrorKind {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "Text parse error: ")?;

        use TextParseErrorKind::*;
        match self {
            InvalidTag => write!(f, "invalid tag"),
            UnexpectedValue => write!(f, "unexpected value"),
            UnexpectedTag => write!(f, "unexpected tag"),
        }
    }
}

#[derive(Clone, Debug)]
pub struct TextBlock {
    pub initial_indent: Option<StyledString>,
    pub subsequent_indent: Option<StyledString>,
    pub text: TaggedText,
}

impl TextBlock {
    // FIXME this function discards tag information
    pub fn layout(&self, base: &TextStyle, width: usize) -> TextLayout {
        let text = self.text.render(&base);

        let initial_indent = self
            .initial_indent
            .as_ref()
            .map(|text| text.render(&base))
            .unwrap_or("".to_string());

        let subsequent_indent = self
            .subsequent_indent
            .as_ref()
            .map(|text| text.render(&base))
            .unwrap_or("".to_string());

        let options = textwrap::Options::new(width)
            .initial_indent(&initial_indent)
            .subsequent_indent(&subsequent_indent);

        TextLayout {
            lines: textwrap::wrap(&text, options)
                .into_iter()
                .map(|line| TaggedText {
                    strings: vec![StyledString::from_plain(&line)],
                })
                .collect(),
            width,
        }
    }
}

#[derive(Clone, Debug)]
pub struct TextLayout {
    pub lines: Vec<TaggedText>,
    pub width: usize,
}

impl TextLayout {
    pub fn from_plain(plain: &str) -> Self {
        let mut lines = Vec::new();
        let mut width = 0;

        for line in plain.lines() {
            width = width.max(textwrap::core::display_width(&line));
            lines.push(TaggedText::from_plain(line));
        }

        Self { lines, width }
    }

    pub fn render(&self, base: &TextStyle) -> RenderedBlock {
        let mut lines = Vec::new();

        for line in self.lines.iter() {
            let mut line = line.render(base);
            let used_width = textwrap::core::display_width(&line);
            let needed_width = self.width - used_width.min(self.width);
            line.extend(std::iter::repeat(' ').take(needed_width));
            lines.push(line);
        }

        RenderedBlock {
            size: Dimensions {
                width: self.width,
                height: lines.len(),
            },
            lines,
        }
    }
}
